@@ -188,6 +188,7 @@ cspi_init_role_table (AccessibleRole *role_table)
   role_table [Accessibility_ROLE_PARAGRAPH] = SPI_ROLE_PARAGRAPH;
   role_table [Accessibility_ROLE_RULER] = SPI_ROLE_RULER;
   role_table [Accessibility_ROLE_APPLICATION] = SPI_ROLE_APPLICATION;
+  role_table [Accessibility_ROLE_AUTOCOMPLETE] = SPI_ROLE_AUTOCOMPLETE;
 
   return TRUE;
 }
@@ -232,7 +233,7 @@ AccessibleRole_getName (AccessibleRole role)
     }
   else
     {
-      return g_strdup ("");
+      return CORBA_string_dup ("");
     }
 }
 
@@ -501,8 +502,6 @@ Accessible_getRoleName (Accessible *obj)
  * Accessible_getStateSet:
  * @obj: a pointer to the #Accessible object on which to operate.
  *
- * Not Yet Implemented.
- *
  * Returns: a pointer to an #AccessibleStateSet representing the object's current state.
  **/
 AccessibleStateSet *
@@ -518,6 +517,8 @@ Accessible_getStateSet (Accessible *obj)
 	  CSPI_OBJREF (obj), cspi_ev ());
   cspi_return_val_if_ev ("getState", NULL);
 
+  cspi_return_val_if_fail (corba_stateset != CORBA_OBJECT_NIL, NULL);
+  cspi_return_val_if_fail (cspi_ping (corba_stateset), NULL);
   corba_seq = Accessibility_StateSet_getStates (corba_stateset, cspi_ev ());
   cspi_return_val_if_ev ("getState", NULL);
 
@@ -986,7 +987,9 @@ cspi_init_relation_type_table (AccessibleRelationType *relation_type_table)
   relation_type_table [Accessibility_RELATION_SUBWINDOW_OF] = SPI_RELATION_SUBWINDOW_OF;
   relation_type_table [Accessibility_RELATION_EMBEDS] = SPI_RELATION_EMBEDS;
   relation_type_table [Accessibility_RELATION_EMBEDDED_BY] = SPI_RELATION_EMBEDDED_BY;
+  relation_type_table [Accessibility_RELATION_POPUP_FOR] = SPI_RELATION_POPUP_FOR;
   relation_type_table [Accessibility_RELATION_LAST_DEFINED] = SPI_RELATION_LAST_DEFINED;
+  return TRUE;
 }
 
 static AccessibleRelationType
@@ -1146,6 +1149,7 @@ spi_state_to_corba (AccessibleState state)
       MAP_STATE (VERTICAL);
       MAP_STATE (VISIBLE);
       MAP_STATE (MANAGES_DESCENDANTS);
+      MAP_STATE (INDETERMINATE);
     default:
       return Accessibility_STATE_INVALID;
   }
